"""
 Module contains functions used in financial regulation model
"""
from math import log
from scipy.stats import norm
import numpy as np

#-----------------------------------------------------------------------------------------                 
def likelihood_matrix(prob_crisis,
                      num_years = 11,
                      start_year = 1,
                      phase_in_years = 3,
                      econ_growth = 0.04):
    """
    This function generates two versions of a likelihood matrix with element
    i, j representing the likelihood that i years ahead, the U.S. economy will
    be in the jth year of a financial crisis.
    
    Parameters:
        prob_crisis: Annual probability of the US economy being in a financial crisis
        num_years: Number of years over which to produce likelihood matrices;
                   default value = 11
        start_year: First year to start calculating crisis likelihoods; default value = 1
        phase_in_years: Number of years over which to phase in crisis probabilities;
                        default value = 3
        econ_growth: Assumed economic growth, which is used to scale crisis_matrix by when
                     producing real_crisis_matrix; default value = 0.04
            
    Returns:
        crisis_matrix: A two-dimensional numpy matrix (num_years, num_yrs) that contains 
        the likelihood that i years ahead the US economy will be in the jth year of a
        financial crisis.
            
        real_crisis_matrix: A two-dimensional numpy matrix (num_years, num_years) that 
        contains the likelihoods in crisis_matrix, but where the ith row are inflated by 
        a growth factor representing increases in the ecnonomy between now and i years 
        from now.
    """

    crisis_matrix = np.zeros((num_years, num_years))
    real_crisis_matrix = np.zeros((num_years, num_years))
    phase_in_rates = np.ones(num_years)

    for yr in range(start_year, start_year + phase_in_years- 1):
        phase_in_rates[yr] = (yr / phase_in_years)

    cum_prob = 1 # initializing the likelihood of not being in a crisis

    for yr in range(start_year, num_years):
        crisis_matrix[yr, 0] = cum_prob - cum_prob * (1 - prob_crisis * phase_in_rates[yr])
        real_crisis_matrix[yr, 0] = crisis_matrix[yr, 0] * (1 + econ_growth) ** (yr - 1)
        
        cum_prob = cum_prob * (1 - prob_crisis * phase_in_rates[yr])

        if yr > 0:
            crisis_matrix[yr, 1 : num_years - 1] =  \
              crisis_matrix[yr - 1, 0 : num_years - 2]
            
            real_crisis_matrix[yr, 1 : num_years - 1] =  \
              real_crisis_matrix[yr - 1, 0 : num_years - 2]
 
    return crisis_matrix, real_crisis_matrix


#-----------------------------------------------------------------------------------------
def mkt_val_deposit_insurance(additional_capital,
                              fail_factor,
                              base_capital = 0.1,
                              base_prob_fail = 0.0034, 
                              drift = 0.01, 
                              loss_rate = 0.2):
    """
    Translate expected losses into the market value of deposit insurance. 
    
    Parameters:
        additional_capital: Additional capital-asset ratio induced by policy change
        fail_factor: Factor to adjust base_prob_fail, resulting from policy change
        base_capital: Average capital-asset ratio;
                      default value = 0.1
        prob_fail: Average annual failure rate of FDIC insured depository institutions;
                   default value = 0.0034
        drift: Excess annual return on assest relative to liabilities; 
               default value = 0.01
        loss_rate: Average ratio of losses to the FDIC relative to bank assets; 
                   default value = 0.2
        
    Returns:
        market_value: The market value of deposit insurance per dollar of assets, in basis
                      points        
    
    """
    capital = base_capital + additional_capital
    prob_fail = base_prob_fail * fail_factor
    
    # K0 is the expected log capital ratio one period ahead
    K0 = log(1 + capital + drift)
    
    # Z is the Z statistic of the normal distribution under the probability of default
    Z = -1.0 * norm.ppf(prob_fail)
    
    # From Z and K0 the program deduces the volatility parameter
    sigma = K0 / Z
    
    # The Market Value is then calculated by excluding the drift
    market_value = (1 - norm.cdf(log(1 + capital) / sigma)) * loss_rate
    
    return market_value * 10000

    