"""
 Module contains the financial regulation Model class.
"""

import pandas as pd 
from math import log, exp

from scenario import Scenario
from current_law import CurrentLaw
from results import Results
from functions import likelihood_matrix, mkt_val_deposit_insurance


class Model():
    """
    The Model class contains the data and functions used to simulate various financial
    regulation policy changes.
    
    Attributes:
        current_law: The Current Law budget and economic projections
        
        mod_crisis: The scenario containing the budget and economic projections under a 
            Moderate (financial) Crisis 
        sev_crisis: The scenario containing the budget and economic projections under a
            Severe (financial) Crisis 

        change_CR: The scenario containing the budget and economic projections under a 
            policy that reduced Capital Requirement by 1 percentage point
        repeal_ATR: The scenario containing the budget and economic projections under a
            policy that repealed the Ability-to-Repay mortgage rules

    Methods:
        __init__()
        make_crisis_dict()
        run_change_CR()
        run_elminate_OLA()
        run_repeal_ATR()
    """
    
    def __init__(self, prob_crisis_parameters):

        self.current_law = CurrentLaw("May 2019 Baseline", "current_law")
        
        self.mod_crisis = Scenario("Moderate Crisis Scenario", "moderate_crisis")
        self.mod_crisis.calc_econ_effects(self.current_law)

        self.sev_crisis = Scenario("Severe Crisis Scenario", "severe_crisis")
        self.sev_crisis.calc_econ_effects(self.current_law)

        self.change_CR = Scenario("Lower Capital Requirements by 1% Point", "lower_CR")
        self.change_CR.calc_econ_effects(self.current_law)

        self.eliminate_OLA = Scenario("Eliminate Orderly Liquidation Authority", "eliminate_OLA")
        self.eliminate_OLA.calc_econ_effects(self.current_law)

        self.repeal_ATR = Scenario("Repeal Ability-to-Repay Rules", "repeal_ATR")
        self.repeal_ATR.calc_econ_effects(self.current_law)

        # Create a dictionary of crisis effects
        self.crisis_dict = self.make_crisis_dict(prob_crisis_parameters)

        
    def make_crisis_dict(self, prob_crisis_parameters):
        """
        Make a dictionary of the crisis effects by prob_crisis, severity ('mod' or 'sev'),
        and budget or econ effects.
        
        This gets called once at model initialization.
        
        Parameters:
            prob_crisis_parameters: A Numpy array of the annual probability of the US 
            economy entering a fiscal crisis
        
        Returns:
            dictionary of crisis effects
        """
        prob_crisis_dict = {}

        for prob_crisis in prob_crisis_parameters:
        
            prob_crisis_dict[prob_crisis] = {}        
            
            # Calculate matrices of the likelihood of financial crises occuring
            #
            # Matrices are square (i, j), lower triangular matrices with the probability  
            # of being in year j of a crisis that started in year i.
            #
            # One matrix is nominal, and the other acccounts for economic growth
            prob_crisis_matrix, prob_crisis_matrix_real = likelihood_matrix(prob_crisis)
    
            #-----------------------------------------------------------------------------
            # Moderate Crisis
            #-----------------------------------------------------------------------------
            prob_crisis_dict[prob_crisis]['mod'] = {}
            
            # Calculate budget effects under current law 
            budget_matrix = prob_crisis_matrix_real[:, 0:10] @ self.mod_crisis.budget_effects.iloc[1:11]
            prob_crisis_dict[prob_crisis]['mod']['budget_effects'] =  \
                pd.DataFrame(data = budget_matrix, 
                             columns = self.mod_crisis.budget_effects.columns)
    
            # Calculate economic effects under current law 
            econ_matrix = prob_crisis_matrix[:, 0:10] @ self.mod_crisis.econ_effects.iloc[1:11]
            prob_crisis_dict[prob_crisis]['mod']['econ_effects'] =  \
                pd.DataFrame(data = econ_matrix,
                             columns = self.mod_crisis.econ_effects.columns)
            
    
            #-----------------------------------------------------------------------------
            # Severe Crisis
            #-----------------------------------------------------------------------------
            prob_crisis_dict[prob_crisis]['sev'] = {}
            
            # Calculate budget effects under current law
            budget_matrix = prob_crisis_matrix_real[:,0:10] @ self.sev_crisis.budget_effects.iloc[1:11]
            prob_crisis_dict[prob_crisis]['sev']['budget_effects'] =  \
                pd.DataFrame(data = budget_matrix,
                             columns = self.sev_crisis.budget_effects.columns)
                
            # Calculate economic effects of under current law 
            econ_matrix = prob_crisis_matrix[:,0:10] @ self.sev_crisis.econ_effects.iloc[1:11]
            
            prob_crisis_dict[prob_crisis]['sev']['econ_effects'] =  \
                pd.DataFrame(data = econ_matrix,
                             columns = self.sev_crisis.econ_effects.columns)
    
        return prob_crisis_dict
    
    #-------------------------------------------------------------------------------------    
    # Change Capital Requirements (CR)
    #-------------------------------------------------------------------------------------
    def run_change_CR(self, params):

        results = Results("Change Capital Requirements (CR)")

        ##--------------------------------------------------------------------------------
        ## Calculate direct budgetary effects        
        ##--------------------------------------------------------------------------------
        
        # Normalize the CR_change policy options relative to the central estimate.
        # For the central estimate, where capital requirements go down by 0.01 this will 
        # equal 1
        CR_change_scalar = params['CR_change'] / (-0.01)
       
        # Because not all institutions covered by Orderly Liquidation Authority
        # are banks, lower capital requirements for banks would raise net outlays from the 
        # Orderly Liquidation Fund (OLA) by a smaller amount than they would for the 
        # Deposit Insurnce Fund (DIF). The factor is based on CBO's models of the cash 
        # flows of the OLF, as estimated by CBO's Budget Analysis Division
        OLA_factor = 0.15

        results.budget['direct_revenues_total'] =  \
            self.current_law.FDIC_vars['OLF_revenue'] / 1000   \
          * OLA_factor  \
          * params['CR_failure']  \
          * CR_change_scalar

        spending_change = (self.current_law.FDIC_vars['OLF_spending'] * OLA_factor  \
                         + self.current_law.FDIC_vars['DIF_spending']) / 1000  \
                        * params['CR_failure']
                        
        premium_change = self.current_law.FDIC_vars['premium_base'] / 1000  \
                       * params['premium_change'] / 10000

        results.budget['direct_outlays_total'] =  \
            (spending_change - premium_change) * CR_change_scalar  

        results.budget['direct_primary_deficit'] =  \
            results.budget['direct_outlays_total'] \
          - results.budget['direct_revenues_total']
          
        ##--------------------------------------------------------------------------------
        ## Calculate noncrisis effects
        ##--------------------------------------------------------------------------------
        
        # Set deposit convenience to 0.3 basis points
        #
        # A recent study suggested that the liquidity that would result from lowering 
        # capital requirements would cut banks’ financing costs by 0.3 basis points per 
        # percentage-point change in the minimum capital levels.
        # See Arvind Krishnamurthy and Annette Vissing-Jorgenson, 
        # “The Aggregate Demand for Treasury Debt,”
        # Journal of Political Economy, vol. 120, no. 2 (April 2012), pp. 233–267, 
        # https://doi.org/10.1086/666526. 
        #
        # That study estimated that the discount on U.S. Treasury bonds compared with 
        # other bonds is 73 basis points, with as much as 46 basis points attirbutable to 
        # liquidity. 
        #
        # CBO estimated the additional convenience value of lowering capital requirements 
        # by 1 percentage points on the basis of that estimate: The agency’s estimate of
        # 0.3 basis points is roughly 1 percent of 46 basis points, adjusted slightly 
        # downward to reflect the fact that only part of the debt that replaced capital
        # would be in the form of deposits. 
        deposit_convenience = 0.3

        # 1 + change in bank failure rates as a result of change in capital
        failure_rate_factor = exp(log(1 + params['CR_failure']) * CR_change_scalar)

        # Take Difference in market value of deposit insurance with change in policy 
        mkt_val_diff =  \
            mkt_val_deposit_insurance(additional_capital = params['CR_change'],
                                      fail_factor = failure_rate_factor) \
          - mkt_val_deposit_insurance(additional_capital = 0,
                                      fail_factor = 1)

        # Calculate the change in financing cost based on four factors
        financing_cost_change = mkt_val_diff  \
                              + (deposit_convenience
                               - params['tax_change']
                               - params['premium_change']) * CR_change_scalar
        
        # Set financing factor to 1.25.
        # To comply with liquidity requirements, banks must invest approximately 20 
        # percent of their assets in risk-free securities (on which they earn interest at 
        # the rates on Treasury securities of similar maturity). They must cover the 
        # difference between their own cost of financing and the Treasury rates by 
        # charging consumers rates that are higher than the Treasury rates on the 
        # remaining 80 percent of their assets. 
        #
        # For every basis point that they need to earn on the entirety of their assets, 
        # therefore, they need to charge 1.25 (1.0 / 0.8) basis points more on that 
        # remaining 80 percent.
        financing_factor = 1.25
        
        # The macro projections for the capital requirements were based on a five basis 
        # point reduction in borrowing costs
        basis_points = -5
        
        # Scale the baseline Macroeconomic and feedback effects based on the calculated
        # fiancing cost change resulting from the CR_change options
        #
        # The macro simulations were based on a 5 basis point change in costs to borrowers
        # The scaling factor equals the ratio of the financing cost to that 5 bp cost to
        # borrowers.
        #
        # For example if the estimated change in financing cost is 7 basis point, the 
        # financing_cost_scalar will equal 1.4
        financing_cost_scalar = (financing_cost_change * -financing_factor) \
                              / basis_points 

        # Scale baseline economic effects to size of effect on financing costs
        results.noncrisis_econ = self.change_CR.econ_effects * financing_cost_scalar 

        # Calculate effect on primary deficit and total deficit
        results.budget['noncrisis_revenues_gdp_growth'] =  \
            self.change_CR.budget_effects['Revenue'] * financing_cost_scalar
       
        results.budget['noncrisis_revenues_financing_mix'] =  \
            self.current_law.FDIC_vars['bank_assets'] / 10000000  \
          * params['tax_change']   \
          * CR_change_scalar
          
        results.budget['noncrisis_revenues_total'] =  \
            results.budget['noncrisis_revenues_gdp_growth']  \
          + results.budget['noncrisis_revenues_financing_mix']

        results.budget['noncrisis_outlays_noninterest'] =  \
            self.change_CR.budget_effects['Noninterest'] * financing_cost_scalar
        
        results.budget['noncrisis_outlays_interest'] =  \
            self.change_CR.budget_effects['Interest'] * financing_cost_scalar

        results.budget['noncrisis_outlays_total'] =  \
            results.budget['noncrisis_outlays_noninterest']  \
          + results.budget['noncrisis_outlays_interest']
                                             
        results.budget['noncrisis_primary_deficit'] =  \
            results.budget['noncrisis_outlays_total']  \
          - results.budget['noncrisis_revenues_total']
          
        
        ##--------------------------------------------------------------------------------
        ## Calculate the crisis effects
        ##--------------------------------------------------------------------------------
        
        # Scale crisis effects to size of effect on failure rates
        results.crisis_econ =  \
            self.crisis_dict[params['prob_crisis']]['mod']['econ_effects']  \
          * (failure_rate_factor - 1)
        
        results.budget['crisis_revenues_total'] = \
            self.crisis_dict[params['prob_crisis']]['mod']['budget_effects']['Revenue'].values  \
          * (failure_rate_factor - 1)
 
        results.budget['crisis_outlays_noninterest'] =  \
            self.crisis_dict[params['prob_crisis']]['mod']['budget_effects']['Noninterest'].values  \
          * (failure_rate_factor - 1)
        
        results.budget['crisis_outlays_interest'] =  \
            self.crisis_dict[params['prob_crisis']]['mod']['budget_effects']['Interest'].values \
          * (failure_rate_factor - 1)
        
        results.budget['crisis_outlays_total'] =  \
            results.budget['crisis_outlays_noninterest']  \
          + results.budget['crisis_outlays_interest']
             
        results.budget['crisis_primary_deficit'] =  \
            results.budget['crisis_outlays_total']  \
          - results.budget['crisis_revenues_total']
             
        return results.budget.sum().to_frame().T
        
    #-------------------------------------------------------------------------------------
    # Eliminate Ordinary Liquidtion Authority (OLA) 
    #-------------------------------------------------------------------------------------        
    def run_eliminate_OLA(self, params):
        
        results = Results("Eliminate Orderly Liquidation Authority (OLA)")
  
        ##--------------------------------------------------------------------------------
        ## Calculate direct budget effects
        ##--------------------------------------------------------------------------------
        # Copy the direct effects of Eliminating OLA from the Current Law FDIC vars
        results.budget['direct_revenues_total'] = self.current_law.FDIC_vars['OLF_revenue'] / 1000
        results.budget['direct_outlays_total'] = self.current_law.FDIC_vars['OLF_spending'] / 1000
        results.budget['direct_primary_deficit'] =  \
            results.budget['direct_outlays_total']  \
          - results.budget['direct_revenues_total']

        ##--------------------------------------------------------------------------------
        ## Calculate noncrisis effects
        ##--------------------------------------------------------------------------------
        # CBO estimates that eliminating OLA will have no budgetary feedback effects under
        # normal (non-crisis) conditions
        results.budget['noncrisis_revenues_gdp_growth'] = 0
        results.budget['noncrisis_revenues_financing_mix'] = 0
        results.budget['noncrisis_revenues_total'] = 0
        results.budget['noncrisis_outlays_noninterest'] = 0
        results.budget['noncrisis_outlays_interest'] = 0
        results.budget['noncrisis_outlays_total'] = 0
        results.budget['noncrisis_primary_deficit'] = 0

        ##--------------------------------------------------------------------------------
        ## Calculate crisis effects 
        ##--------------------------------------------------------------------------------
        results.crisis_econ =  \
            (self.crisis_dict[params['prob_crisis']]['sev']['econ_effects']  \
           - self.crisis_dict[params['prob_crisis']]['mod']['econ_effects'])  \
            * params['OLA_fraction']  \
            * (1 - params['OLA_effect'])  \
            + self.crisis_dict[params['prob_crisis']]['mod']['econ_effects']  \
            * -params['OLA_effect']

        results.crisis_budget =  \
            (self.crisis_dict[params['prob_crisis']]['sev']['budget_effects']  \
           - self.crisis_dict[params['prob_crisis']]['mod']['budget_effects'])  \
            * params['OLA_fraction']  \
            * (1 - params['OLA_effect'])  \
            + self.crisis_dict[params['prob_crisis']]['mod']['budget_effects']  \
            * -params['OLA_effect']
        
        results.budget['crisis_revenues_total'] = results.crisis_budget['Revenue'].values
        results.budget['crisis_outlays_noninterest'] = results.crisis_budget['Noninterest'].values
        results.budget['crisis_outlays_interest'] = results.crisis_budget['Interest'].values
        results.budget['crisis_outlays_total'] = results.crisis_budget['Noninterest'].values  \
                                               + results.crisis_budget['Interest'].values
        
        results.budget['crisis_primary_deficit'] =  \
            results.budget['crisis_outlays_total']  \
          - results.budget['crisis_revenues_total']
        
        return results.budget.sum().to_frame().T
     
        
    #-------------------------------------------------------------------------------------
    # Repeal Ability-to-Repay (ATR) mortgage rule
    #-------------------------------------------------------------------------------------     
    def run_repeal_ATR(self, params):
        
        results = Results("Repeal Ability-to-Repay (ATR) mortgage rule")
 
        ##--------------------------------------------------------------------------------
        ## Calculate direct budgetary effects
        ##--------------------------------------------------------------------------------
        
        # Revenues: Orderly Liquidation Fund (OLF) revenues increase by the increased 
        # failure rate in the ATR_effect.
        #
        # Spending: Orderly Liquidation Fund (OLF) and the Deposit Insurance Fund (DIF) 
        # spending by the increased failure rate captured in the ATR_effect.
        results.budget['direct_revenues_total'] =  \
            (self.current_law.FDIC_vars['OLF_revenue']) / 1000  \
            * params['ATR_effect']

        results.budget['direct_outlays_total'] =  \
            (self.current_law.FDIC_vars['OLF_spending']  \
           + self.current_law.FDIC_vars['DIF_spending']) / 1000  \
            * params['ATR_effect']
        
        results.budget['direct_primary_deficit'] =  \
            results.budget['direct_outlays_total']  \
          - results.budget['direct_revenues_total']
    
        ##--------------------------------------------------------------------------------
        ## Calculate noncrisis effects
        ##--------------------------------------------------------------------------------
        
        # The repeal ATR macroeconomic projections were developed based on a 5 percent
        # change in origination volume. The policy of repealing Ability to Repay
        # rule, however, was estimated to raise origination volumes by 2 percent. 
        # The factor scales the macroecnomic projections to be consistent with that 
        # increase
        origination_factor = 0.4

        # The revenue for repealling ATR is based on the change in GDP times the effective
        # tax rate on income from owner occupied housing. The revenues are then scaled to
        # account for the different origination rates between the macroeconomic input and
        # the policy.
        results.budget['noncrisis_revenues_total'] =  \
            (self.repeal_ATR.econ['GDP'] - self.current_law.econ['GDP'])  \
            * params['housing_tax']  \
            * origination_factor

        results.budget['noncrisis_outlays_noninterest'] = 0
        results.budget['noncrisis_outlays_interest'] = 0
        results.budget['noncrisis_outlays_total'] = 0
        
        results.budget['noncrisis_primary_deficit'] =  \
            results.budget['noncrisis_outlays_total']  \
          - results.budget['noncrisis_revenues_total']

        ##--------------------------------------------------------------------------------
        ## Calculate crisis effects
        ##--------------------------------------------------------------------------------
        
        # Scale moderate crisis effects based on effect of ATR on failure rates
        results.budget['crisis_revenues_total'] = \
            self.crisis_dict[params['prob_crisis']]['mod']['budget_effects']['Revenue'].values  \
          * params['ATR_effect']
 
        results.budget['crisis_outlays_noninterest'] =  \
            self.crisis_dict[params['prob_crisis']]['mod']['budget_effects']['Noninterest'].values  \
          * params['ATR_effect']
        
        results.budget['crisis_outlays_interest'] =  \
            self.crisis_dict[params['prob_crisis']]['mod']['budget_effects']['Interest'].values \
          * params['ATR_effect']
        
        results.budget['crisis_outlays_total'] =  \
            results.budget['crisis_outlays_noninterest']  \
          + results.budget['crisis_outlays_interest']
        
        results.budget['crisis_primary_deficit'] = results.budget['crisis_outlays_total']  \
                                                 - results.budget['crisis_revenues_total']
        
        return results.budget.sum().to_frame().T
